from typing import Any

from mypy_boto3_rds import RDSClient

from ei.core.service import BaseAwsService


class AwsRdsInstanceService(BaseAwsService):
    service_name = 'rds'

    @classmethod
    def _list(cls, client: RDSClient) -> Any:
        return client.describe_db_instances()['DBInstances']

    @classmethod
    def _show(cls, client: RDSClient, id: str) -> Any:
        return client.describe_db_instances(
            DBInstanceIdentifier=id)['DBInstances']



from typing import Callable
from concurrent.futures import ThreadPoolExecutor
from itertools import chain

from ei.core import defaults


def bulk_action(fn: Callable, regions: list, account_ids: list) -> chain:
    tasks = []
    with ThreadPoolExecutor(max_workers=defaults.CORES) as executor:
        for account_id in account_ids:
            for region in regions:
                tasks.append(executor.submit(fn, region, account_id))

    results = [t.result() for t in tasks]
    iterable = chain(*results)

    return iterable



import typing

from ei.core.cli import create_application
from ei.views.cli import ec2
from ei.views.cli import elasticache
from ei.views.cli import rds
from ei.views.cli import s3
from ei.views.cli import elb


if typing.TYPE_CHECKING:
    from ei.core.cli import Typeable  # noqa: F401


APPS = [
    ec2.group,
    elasticache.group,
    rds.group,
    s3.group,
    elb.group
]  # type: typing.List[Typeable]


cli = create_application(APPS)



from typing import Any

from mypy_boto3_elbv2 import ElasticLoadBalancingv2Client

from ei.core.service import BaseAwsService


class AwsElbLoadbalancerService(BaseAwsService):
    service_name = 'elbv2'

    @classmethod
    def _list(cls, client: ElasticLoadBalancingv2Client) -> Any:
        loadbalancers = client.describe_load_balancers()['LoadBalancers']
        return loadbalancers

    @classmethod
    def _show(cls, client: ElasticLoadBalancingv2Client, id: str) -> Any:
        """Show Loadbalancer Object

        Fetch loadbalancer object and additional info.

        Additional info.

        - [green]Attributes[/green]: boto3.describe_load_balancer_attributes()
        - [green]Listener[/green]: boto3.describe_listeners()
        - [green]TargetGroups[/green]: boto3.describe_target_groups()
        - [green]Tags[/green]: boto3.describe_tags()
        """

        loadbalancer = client.describe_load_balancers(
            LoadBalancerArns=[id])['LoadBalancers']
        attributes = client.describe_load_balancer_attributes(
            LoadBalancerArn=id)
        listeners = client.describe_listeners(LoadBalancerArn=id)
        target_groups = client.describe_target_groups(LoadBalancerArn=id)
        tags = client.describe_tags(
            ResourceArns=[id])['TagDescriptions'][0]['Tags']

        return [{
            **loadbalancer[0],
            'Attributes': attributes.get('Attributes'),
            'Listeners': listeners.get('Listeners'),
            'TargetGroups': target_groups.get('TargetGroups'),
            'Tags': tags
        }]


class AwsElbListenerService(BaseAwsService):
    service_name = 'elbv2'

    @classmethod
    def _list(cls, client: ElasticLoadBalancingv2Client) -> Any:
        listeners = client.describe_listeners()['Listeners']
        return listeners

    @classmethod
    def _show(cls, client: ElasticLoadBalancingv2Client, id: str) -> Any:
        listener = client.describe_listeners(ListenerArns=[id])['Listeners']
        certificates = client.describe_listener_certificates(ListenerArn=id)
        rules = client.describe_rules(ListenerArn=id)

        tags = client.describe_tags(
            ResourceArns=[id])['TagDescriptions'][0]['Tags']

        return [{
            **listener[0],
            'Certificates': certificates.get('Certificates'),
            'Rules': rules.get('Rules'),
            'Tags': tags
        }]


class AwsElbTargetGroupService(BaseAwsService):
    service_name = 'elbv2'

    @classmethod
    def _list(cls, client: ElasticLoadBalancingv2Client) -> Any:
        target_groups = client.describe_target_groups()['TargetGroups']
        return target_groups

    @classmethod
    def _show(cls, client: ElasticLoadBalancingv2Client, id: str) -> Any:
        target_group = client.describe_target_groups(
            TargetGroupArns=[id])['TargetGroups']

        attributes = client.describe_target_group_attributes(
            TargetGroupArn=id)
        health = client.describe_target_health(TargetGroupArn=id)
        tags = client.describe_tags(
            ResourceArns=[id])['TagDescriptions'][0]['Tags']

        return [{
            **target_group[0],
            'Certificates': attributes.get('Attributes'),
            'Rules': health.get('TargetHealthDescriptions'),
            'Tags': tags
        }]



import os


CORES = os.cpu_count()

CONFIGS = (
    EI_CREDENTIAL_RESOLVER,

    EI_ACCOUNT_IDS,

    # use all regions
    # from botocore.args import LEGACY_GLOBAL_STS_REGIONS
    EI_REGIONS,

    # sts credential resolver config, if credential resolver is 'env', then
    # this configurations will be ignored.
    EI_STS_ASSUME_ROLE_ARN_PATTERN,
    EI_STS_ASSUME_ROLE_SESSION_NAME,

    AWS_REGION,
    AWS_ACCESS_KEY_ID,
    AWS_SECRET_ACCESS_KEY,
    AWS_SECURITY_TOKEN,
    AWS_SESSION_EXPIRATION,
) = (
    os.getenv('EI_CREDENTIAL_RESOLVER', 'sts'),

    os.getenv('EI_ACCOUNT_IDS', '').split(','),
    os.getenv('EI_REGIONS', '').split(','),
    os.getenv('EI_STS_ASSUME_ROLE_ARN_PATTERN', ''),
    os.getenv('EI_STS_ASSUME_ROLE_SESSION_NAME', ''),
    os.getenv('AWS_REGION', ''),
    os.getenv('AWS_ACCESS_KEY_ID', ''),
    os.getenv('AWS_SECRET_ACCESS_KEY', ''),
    os.getenv('AWS_SECURITY_TOKEN', ''),
    os.getenv('AWS_SESSION_EXPIRATION', '')
)


def print_config() -> str:
    return '\n'.join((
        f'EI_CREDENTIAL_RESOLVER={EI_CREDENTIAL_RESOLVER!r}',
        f'EI_ACCOUNT_IDS={EI_ACCOUNT_IDS}',
        f'EI_REGIONS={EI_REGIONS}',
        ('EI_STS_ASSUME_ROLE_ARN_PATTERN='
            f'{EI_STS_ASSUME_ROLE_ARN_PATTERN!r}'),
        ('EI_STS_ASSUME_ROLE_SESSION_NAME='
            f'{EI_STS_ASSUME_ROLE_SESSION_NAME!r}'),
        f'AWS_REGION={AWS_REGION!r}',
        f'AWS_ACCESS_KEY_ID={AWS_ACCESS_KEY_ID!r}',
        f'AWS_SECRET_ACCESS_KEY={AWS_SECRET_ACCESS_KEY!r}',
        f'AWS_SECURITY_TOKEN={AWS_SECURITY_TOKEN!r}',
        f'AWS_SESSION_EXPIRATION={AWS_SESSION_EXPIRATION!r}',
    ))



from contextlib import _GeneratorContextManager
from typing import Any
from typing import List
from typing import Callable

from botocore.client import BaseClient

from ei.core import defaults
from ei.core.session import client_session


class BaseAwsService(object):
    service_name: str
    _sessioncontext: Callable[..., _GeneratorContextManager] = client_session

    @classmethod
    def _list(cls, client: BaseClient) -> Any:
        """List diapatcher using client

        Should returns aws resource list

        Example =
            [
                {object A},
                {object B},
                {object C},
                ...
            ]
        """

        raise NotImplementedError('_list() dispatcher not implemented')

    @classmethod
    def _show(cls, client: BaseClient, id: str) -> Any:
        """Show diapatcher using aws client

        Should returns aws resource

        Example =
            {object A}
        """

        raise NotImplementedError('_show() dispatcher not implemented')

    @classmethod
    def _defaulting_args(cls, region: str, account_id: str) -> tuple:
        if not region:
            region = defaults.AWS_REGION

        if not account_id:
            account_id = defaults.EI_ACCOUNT_IDS[0]

        return region, account_id

    @classmethod
    def _dispatch(cls, fn: Callable, region: str, account_id: str,
                  *args: Any, **kwargs: Any) -> Any:
        """Dispatching resources by command
        """

        result = None

        with cls._sessioncontext(
                account_id=account_id, region=region,
                service_name=cls.service_name) as client:

            result = fn(client, *args, **kwargs)
            result = [
                {'Region': region, 'Account': account_id, **obj}
                for obj in result
            ]

        return result

    @classmethod
    def list(cls, region: str, account_id: str) -> List[Any]:
        region, account_id = cls._defaulting_args(region, account_id)
        result = cls._dispatch(
            cls._list, region=region, account_id=account_id)  # type: list

        return result

    @classmethod
    def show(cls, id: str, region: str, account_id: str) -> Any:
        region, account_id = cls._defaulting_args(region, account_id)
        result = cls._dispatch(
            cls._show, region=region, account_id=account_id, id=id)[0]

        return result



from typing import Type

from ei.core.cli import BaseCliApp
from ei.core.cli import CliGroup
from ei.core.service import BaseAwsService
from ei.core.fields import Field
from ei.core.fields import IDField
from ei.core.fields import BooleanField
from ei.core.fields import TagField
from ei.core.fields import DictField
from ei.core.fields import extract_from_tag
from ei.core.fields import extract
from ei.services.aws.ec2 import AwsEc2VpcService
from ei.services.aws.ec2 import AwsEc2SubnetService
from ei.services.aws.ec2 import AwsEc2InstanceService
from ei.services.aws.ec2 import AwsEc2AmiService


group = CliGroup(name='ec2', description='AWS EC2')


@group.app
class Ec2VpcCli(BaseCliApp):
    name: str = 'vpc'
    description: str = 'EC2 VPC'

    service_cls = AwsEc2VpcService

    short_fields = (
        IDField('VpcId'),
        Field('Name', serializer=extract_from_tag('Name')),
        Field('InstanceTenancy'),
        BooleanField('IsDefault'),
        Field('State'),
        Field('DhcpOptionsId'),
        Field('CidrBlock')
    )

    long_fields = (
        DictField('CidrBlockAssociationSet'),
        TagField('Tags')
    )


@group.app
class Ec2SubnetCli(BaseCliApp):
    name: str = 'subnet'
    description: str = 'EC2 Subnet'
    service_cls = AwsEc2SubnetService

    short_fields = (
        IDField('SubnetId'),
        Field('Name', serializer=extract_from_tag('Name')),
        Field('CidrBlock'),
        Field('AvailableIpAddressCount'),
        BooleanField('DefaultForAz'),
        BooleanField('MapPublicIpOnLaunch'),
        Field('State'),
        Field('VpcId'),
    )

    long_fields = (
        BooleanField('AssignIpv6AddressOnCreation'),
        DictField('Ipv6CidrBlockAssociationSet'),
        TagField('Tags')
    )

    detail_fields = (
        Field('SubnetArn'),
        Field('OutpostArn'),
        Field('OwnerId'),
        Field('AvailabilityZone'),
        Field('AvailabilityZoneId'),
        Field('EnableLniAtDeviceIndex'),
        BooleanField('EnableDns64'),
        BooleanField('Ipv6Native'),
        Field('CustomerOwnedIpv4Pool'),
        BooleanField('MapCustomerOwnedIpOnLaunch'),
        DictField('PrivateDnsNameOptionsOnLaunch'),
    )


@group.app
class Ec2Instance(BaseCliApp):
    name: str = 'instance'
    description: str = 'EC2 instance'

    service_cls: Type[BaseAwsService] = AwsEc2InstanceService

    stats_fields = ['Region', 'Account', 'InstanceType', 'ImageId', 'State']

    short_fields = (
        IDField('InstanceId'),
        Field('Name', serializer=extract_from_tag('Name')),
        Field('ImageId'),
        Field('InstanceType'),
        Field('PrivateIpAddress'),
        Field('State', serializer=extract('Name')),
        Field('VpcId'),
    )

    long_fields = (
        Field('SubnetId'),
        Field('Hypervisor'),
        Field('Architecture'),
        TagField('Tags')
    )

    detail_fields = (
        Field('AmiLaunchIndex'),
        Field('LaunchTime'),
        Field('PrivateDnsName'),
        Field('ProductCodes'),
        Field('PublicDnsName'),
        Field('StateTransitionReason'),
        Field('ClientToken'),
        BooleanField('EbsOptimized'),
        BooleanField('EnaSupport'),
        DictField('Placement'),
        DictField('BlockDeviceMappings'),
        DictField('MaintenanceOptions'),
        DictField('Monitoring'),
        DictField('IamInstanceProfile'),
        DictField('NetworkInterfaces'),
        Field('RootDeviceName'),
        Field('RootDeviceType'),
        DictField('SecurityGroups'),
        BooleanField('SourceDestCheck'),
        Field('VirtualizationType'),
        DictField('CpuOptions'),
        DictField('CapacityReservationSpecification'),
        DictField('HibernationOptions'),
        DictField('MetadataOptions'),
        DictField('EnclaveOptions'),
        Field('PlatformDetails'),
        Field('UsageOperation'),
        Field('UsageOperationUpdateTime'),
        DictField('PrivateDnsNameOptions'),
    )


@group.app
class Ec2Ami(BaseCliApp):
    name: str = 'ami'
    description: str = 'EC2 AMI'

    service_cls: Type[BaseAwsService] = AwsEc2AmiService

    short_fields = (
        IDField('ImageId'),
        Field('Name'),
        BooleanField('Public'),
        Field('State'),
        Field('CreationDate'),
    )

    long_fields = (
        Field('Description'),
        Field('ImageType'),
        Field('Hypervisor'),
        Field('Architecture'),
        Field('PlatformDetails'),
        TagField('Tags')
    )

    detail_fields = (
        Field('UsageOperation'),
        Field('RootDeviceName'),
        Field('RootDeviceType'),
        Field('EnaSupport'),
        Field('SriovNetSupport'),
        Field('VirtualizationType'),
        Field('DeprecationTime'),
        Field('ImageLocation'),
        Field('OwnerId'),
        DictField('BlockDeviceMappings'),
        Field('ImageOwnerAlias'),
    )



from collections import defaultdict
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Iterable
from typing import Optional

from typer import Exit
from typer import Typer
from rich.console import Console
from botocore.exceptions import ClientError

from ei.__about__ import VERSION
from ei.core import defaults
from ei.core.fields import Field
from ei.core.service import BaseAwsService
from ei.core.concurrency import bulk_action
from ei.core.exceptions import BaseError
from ei.core.exceptions import PreflightError
from ei.core.exceptions import WrongRegionError
from ei.core.exceptions import WrongAccountError
from ei.core.table import list_table
from ei.core.table import detail_table


TOPK = 10


def create_application(apps: List['Typeable']) -> Typer:
    cli = CliGroup(
        name='ei',
        description=(
            'A[bold green](ei)[/bold green]'
            f' - AWS CLI for humans. ({VERSION})\n\n'
        ),
        apps=apps
    )

    typer = cli.typer()

    @typer.command(help='Show Current Configuration')
    def config() -> None:
        console = Console()
        console.print(defaults.print_config())

    return typer


def _get_typer(name: str, help: str) -> Typer:
    app = Typer(
        name=name,
        help=help,
        no_args_is_help=True,
        rich_markup_mode='rich',
    )
    return app


def _preflight() -> None:
    if not all(defaults.CONFIGS):
        raise PreflightError((
            'Environment variables are not fulfilled.'
            ' Check the environment variables.'
        ))


def _serialize_data_as_list(fields: Iterable[Field],
                            results: Iterable) -> list:
    return [
        [
            serializer(item)
            for serializer in fields
        ] for item in results
    ]


def _serialize_data_as_dict(fields: Iterable[Field], result: Dict) -> Dict:
    serialized = {}

    for field in fields:
        serialized[field._name] = field(result)

    return serialized


class Typeable(object):
    name: str
    description: str = ''

    def typer(self) -> Typer:
        raise NotImplementedError()


def guard_common_errors(fn: Callable) -> Callable:
    """Wrap exposing command function with common Exception handlers

    Common Exceptions:

    - PreflightError: failed to check default configuration (env vars)
    - BaseError, ClientError: Expected Exception from AWS or client
    - NotImplementedError: Not Supported command
    """

    @wraps(fn)
    def _wraps(self: 'BaseCliApp', region: str = '', account_id: str = '',
               *args: tuple, **kwargs: Dict) -> Any:
        try:
            _preflight()
            self._validate_region(region)
            self._validate_acocunt_id(account_id)

        except PreflightError as e:
            self._console.print(e, style='red')
            self._console.print(f'\n{defaults.print_config()}')
            raise Exit(code=1)

        try:
            result = fn(
                self, region=region, account_id=account_id, *args, **kwargs)

            return result

        except (BaseError, ClientError) as e:
            print(e)
            raise Exit(code=1)

        except NotImplementedError:
            self._console.print(
                '[yellow]This command is not supported yet. :grin:[/yellow]')
            raise Exit(code=1)

    return _wraps


class BaseCliApp(Typeable):
    service_cls: Type[BaseAwsService]

    # fields for list
    short_fields: tuple

    # additional fields for long list
    long_fields: tuple

    # more additional fields for show resource
    detail_fields: tuple = ()

    stats_fields = ['Region', 'Account']

    _service: BaseAwsService

    def __init__(self) -> None:
        self._service = self.service_cls()
        self._console = Console()

        self._list_detail_fields = self.short_fields + self.long_fields
        self._full_fields = self._list_detail_fields + self.detail_fields

    def _validate_region(self, region: str) -> None:
        if region and region not in defaults.EI_REGIONS:
            raise WrongRegionError(
                f'region "{region}" is not in "EI_REGIONS"'
                f'(EI_REGIONS={defaults.EI_REGIONS})'
            )

    def _validate_acocunt_id(self, account_id: str) -> None:
        if account_id and account_id not in defaults.EI_ACCOUNT_IDS:
            raise WrongAccountError(
                f'account_id "{account_id}" is not in "EI_ACCOUNT_IDS"'
                f'(EI_ACCOUNT_IDS={defaults.EI_ACCOUNT_IDS})'
            )

    def _defaulting_args(
            self,
            long: bool,
            region: str,
            all_regions: bool,
            account_id: str,
            all_accounts: bool) -> tuple:

        if long:
            fields = self._list_detail_fields
        else:
            fields = self.short_fields

        additional_fields = []
        regions = defaults.EI_REGIONS
        if not all_regions:
            regions = [region]
        else:
            additional_fields.append(Field('Region'))

        account_ids = defaults.EI_ACCOUNT_IDS
        if not all_accounts:
            account_ids = [account_id]
        else:
            additional_fields.append(Field('Account'))

        display_fields = additional_fields + [*fields]

        return (regions, account_ids, display_fields)

    @guard_common_errors
    def list(
            self,
            long: bool = False,
            stat: bool = True,
            table: bool = True,
            region: str = '',
            account_id: str = '',
            all_regions: bool = False,
            all_accounts: bool = False) -> int:
        """List resources
        """

        regions, account_ids, display_fields = self._defaulting_args(
            long, region, all_regions, account_id, all_accounts)

        results = [*bulk_action(self._service.list, regions, account_ids)]
        serialized_results = _serialize_data_as_list(
            display_fields, results)

        if table:
            display_table = list_table([
                field._name for field in display_fields
            ], serialized_results)
            self._console.print(display_table)

        if stat:
            stats_dict: Dict = {}
            for subject in self.stats_fields:
                stats_dict[subject] = defaultdict(int)

                for item in results:
                    stats_dict[subject][
                        str(item.get(subject))] += 1

            self._console.print(
                f'{len(serialized_results)} "{self.name}" items.')

            for subject in self.stats_fields:
                results = list(stats_dict[subject].items())
                results.sort(
                    key=lambda x: x[1], reverse=True)  # type: ignore

                stat_header = f'\n* per {subject}'
                if len(results) > TOPK:
                    stat_header += (
                        f' (top:{TOPK} / total categories:{len(results)})')

                self._console.print(stat_header)
                for key, count in results[:TOPK]:
                    self._console.print(f'  - "{key}": {count} items.')

        return 0

    @guard_common_errors
    def show(
            self,
            id: str,
            region: str = '',
            account_id: str = '') -> int:
        """Show resource
        """

        result = self._service.show(id, region, account_id)
        serialized_result = _serialize_data_as_dict(
            self._full_fields, result)

        table = detail_table(serialized_result)
        self._console.print(table)

        return 0

    def typer(self) -> Typer:
        commands = ['list', 'show']  # type: List[str]
        available_commands = ', '.join([
            f'[bright_blue]{cmd}[/bright_blue]'
            for cmd in commands
        ])

        app = _get_typer(
            name=self.name,
            help=(
                f'{self.description} [bright_black]'
                f'(subcommands: {available_commands})'
                '[/bright_black]'
            )
        )

        for cmd in commands:
            fn = getattr(self, cmd)
            service_fn = getattr(self.service_cls, f'_{cmd}')
            # replace help message to implemented function's docstring.
            app.command(help=service_fn.__doc__)(fn)

        return app


class CliGroup(Typeable):
    _apps: List[Typeable]

    def __init__(
            self,
            name: str,
            description: str,
            apps: Optional[List[Typeable]] = None
            ) -> None:

        self.name = name
        self.description = description
        self._apps = []

        if apps:
            for app in apps:
                self.add(app)

    def app(self, cls: Type[Typeable]) -> None:
        instance = cls()
        self.add(instance)

    def add(self, instance: Typeable) -> None:
        self._apps.append(instance)

    def typer(self) -> Typer:
        group_description = ', '.join([
            f'[bright_blue]{app.name}[/bright_blue]'
            for app in self._apps
        ])

        group = _get_typer(
            name=self.name,
            help=(
                f'{self.description} [bright_black]'
                f'(subcommands: {group_description})'
                '[/bright_black]'
            )
        )

        for app in self._apps:
            group.add_typer(app.typer())

        return group



import pytest
from ei.core.fields import extract
from ei.core.fields import extract_from_tag
from ei.core.fields import Field
from ei.core.fields import DictField
from ei.core.fields import IDField
from ei.core.fields import BooleanField
from ei.core.fields import TagField


@pytest.mark.parametrize(['field', 'obj', 'expected'], argvalues=[
    ('field-a', {'field-a': 'expected'}, 'expected'),
    ('field-a', {'field-b': 'not expected'}, ''),
])
def test_field(field: str, obj: dict, expected: str) -> None:
    serializer = Field(field)
    value = serializer(obj)
    assert value == expected


@pytest.mark.parametrize(['field', 'obj', 'expected'], argvalues=[
    ('true-key', {'true-key': True}, '[green]True[/green]'),
    ('false-key', {'false-key': False}, '[bright_red]False[/bright_red]'),
])
def test_booleanfield(field: str, obj: dict, expected: str) -> None:
    serializer = BooleanField(field)
    true_value = serializer(obj)
    assert true_value == expected


@pytest.mark.parametrize(['field', 'key', 'obj', 'expected'], argvalues=[
    ('dict', 'key', {'dict': {'key': 'expected'}}, 'expected'),
    ('dict', 'no-key', {'dict': {'key': 'expected'}}, ''),
])
def test_field_with_extract(field: str, key: str,
                            obj: dict, expected: str) -> None:
    serializer = Field(field, serializer=extract(key))
    value = serializer(obj)
    assert value == expected


@pytest.mark.parametrize(['field', 'key', 'obj', 'expected'], argvalues=[
    ('tags', 'name', {'Tags': [{'Key': 'name', 'Value': 'john'}]}, 'john'),
    ('tags', 'no-key', {'Tags': [{'Key': 'name', 'Value': 'john'}]}, ''),
    ('tags', 'no-key', {'no-tag': ''}, ''),
])
def test_field_with_extract_from_tag(field: str, key: str,
                                     obj: dict, expected: str) -> None:
    serializer = Field(field, serializer=extract_from_tag(key))
    value = serializer(obj)
    assert value == expected


def test_extract_should_returns_extracted_value_from_dict() -> None:
    serializer = extract('expected-key')
    value = serializer({}, {'expected-key': 'expected-value'})
    assert value == 'expected-value'


def test_idfield_should_returns_formatted_string() -> None:
    field = IDField('id')
    value = field({'id': 1})
    assert value == '[bold]1[/bold]'


def test_tagfield_should_returns_formatted_string() -> None:
    serializer = TagField('Tags')
    value = serializer({
        'Tags': [
            {'Key': 'name', 'Value': 'john'},
            {'Key': 'species', 'Value': 'human'}
        ]
    })

    assert value == (
        '[bright_black]name[/bright_black]: john\n'
        '[bright_black]species[/bright_black]: human'
    )

    value = serializer({'Tags': []})

    assert value == ''


def test_dictfield_should_returns_formatted_string() -> None:
    serializer = DictField('Object')
    value = serializer({
        'Object': {
            'Key': 'Value'
        }
    })

    assert value == "[bright_black]Key[/bright_black]: 'Value'"


def test_dictfield_should_returns_formatted_string_when_empty_value() -> None:
    serializer = DictField('Object')
    value = serializer({
        'Object': None
    })

    assert value == '[bright_black]None[/bright_black]'


def test_dictfield_should_ignores_response_metadata() -> None:
    serializer = DictField('Object')
    value = serializer({
        'Object': {
            'Key': 'Value'
        },
        'ResponseMetadata': {}
    })

    assert value == "[bright_black]Key[/bright_black]: 'Value'"


def test_dictfield_should_returns_if_dictlist() -> None:
    serializer = DictField('Object')
    value = serializer({
        'Object': [{
            'KeyA': 'ValueA'
        }, {
            'KeyB': 'ValueB'
        }],
        'ResponseMetadata': {}
    })

    assert value == '\n'.join([
        "- [bright_black]KeyA[/bright_black]: 'ValueA'",
        "- [bright_black]KeyB[/bright_black]: 'ValueB'"
    ])



import typing

from ei.core.cli import create_application
from ei.views.cli import ec2
from ei.views.cli import elasticache
from ei.views.cli import rds


if typing.TYPE_CHECKING:
    from ei.core.cli import Typeable  # noqa: F401


def test_create_application() -> None:
    apps = [
        ec2.group,
        elasticache.group,
        rds.group
    ]  # type: typing.List[Typeable]

    create_application(apps)


def test_main() -> None:
    from ei.main import cli  # noqa: F401



from contextlib import contextmanager
from contextlib import _GeneratorContextManager
from typing import Any
from typing import Callable
from typing import Generator

from ei.core.cli import BaseCliApp
from ei.core.fields import Field
from ei.core.service import BaseAwsService


@contextmanager
def with_dummy(account_id: str, region: str, service_name: str) -> Generator:
    yield ''


ITEMS = [
    {'name': 'item-a', 'type': 'a'},
    {'name': 'item-b', 'type': 'a'},
    {'name': 'item-c', 'type': 'b'},
]


class DummyAwsService(BaseAwsService):
    service_name: str = 'dummy'
    _sessioncontext: Callable[..., _GeneratorContextManager] = with_dummy

    @classmethod
    def _list(cls, client: Any) -> Any:
        return ITEMS

    @classmethod
    def _show(cls, client: Any, id: str) -> Any:
        found_item = [item for item in ITEMS if item['name'] == id]
        return found_item


class DummyAwsCli(BaseCliApp):
    name = 'dummy'

    service_cls = DummyAwsService
    stats_fields = ['type']
    short_fields = (
        Field('name'),
    )

    long_fields = ()



from itertools import chain
from typing import Any

from mypy_boto3_elasticache.client import ElastiCacheClient

from ei.core.service import BaseAwsService


class AwsElasticacheReplicationGroupService(BaseAwsService):
    service_name = 'elasticache'

    @classmethod
    def _list(cls, client: ElastiCacheClient) -> Any:
        groups = client.describe_replication_groups()['ReplicationGroups']
        return groups

    @classmethod
    def _show(cls, client: ElastiCacheClient, id: str) -> Any:
        group = client.describe_replication_groups(
            ReplicationGroupId=id)['ReplicationGroups']

        return group


class AwsElasticacheCacheClusterService(BaseAwsService):
    service_name = 'elasticache'

    @classmethod
    def _list(cls, client: ElastiCacheClient) -> Any:
        paginator = client.get_paginator(
            operation_name='describe_cache_clusters')
        clusters = chain(*[
            page['CacheClusters']
            for page in paginator.paginate()
        ])
        return clusters

    @classmethod
    def _show(cls, client: ElastiCacheClient, id: str) -> Any:
        clusters = client.describe_cache_clusters(
            CacheClusterId=id)['CacheClusters']

        return clusters


class AwsElasticacheEventService(BaseAwsService):
    service_name = 'elasticache'

    @classmethod
    def _list(cls, client: ElastiCacheClient) -> Any:
        paginator = client.get_paginator(operation_name='describe_events')
        events = chain(*[
            page['Events']
            for page in paginator.paginate(Duration=60 * 24)
        ])
        return events

    # @classmethod
    # def _show(cls, client: ElastiCacheClient, id: str) -> Any:
    #     clusters = client.describe_cache_clusters(
    #         CacheClusterId=id)['CacheClusters']

    #     return clusters



from typing import Type
from ei.core.cli import BaseCliApp
from ei.core.cli import CliGroup
from ei.core.service import BaseAwsService
from ei.core.fields import Field
from ei.core.fields import IDField
from ei.core.fields import BooleanField
from ei.core.fields import TagField
from ei.core.fields import DictField
from ei.services.aws.elasticache import AwsElasticacheReplicationGroupService
from ei.services.aws.elasticache import AwsElasticacheCacheClusterService
from ei.services.aws.elasticache import AwsElasticacheEventService


group = CliGroup(name='elasticache', description='AWS Elasticache')


@group.app
class ElasticacheReplicationGroupCli(BaseCliApp):
    name: str = 'replicationgroup'
    description: str = 'Elasticache replication group'

    service_cls: Type[BaseAwsService] = AwsElasticacheReplicationGroupService

    stats_fields = [
        'Region', 'Account', 'ClusterEnabled', 'CacheNodeType', 'Status']

    short_fields = (
        IDField('ReplicationGroupId'),
        Field('Status'),
        Field('AutomaticFailover'),
        Field('MultiAZ'),
        BooleanField('ClusterEnabled'),
        Field('CacheNodeType'),
    )

    long_fields = (
        BooleanField('AuthTokenEnabled'),
        BooleanField('TransitEncryptionEnabled'),
        BooleanField('AtRestEncryptionEnabled'),
        BooleanField('AutoMinorVersionUpgrade'),
        Field('ReplicationGroupCreateTime'),
        Field('SnapshotWindow'),
        Field('SnapshotRetentionLimit'),
        DictField('ConfigurationEndpoint'),
        Field('PendingModifiedValues'),
        Field('LogDeliveryConfigurations'),
        TagField('Tags')
    )

    detail_fields = (
        Field('NetworkType'),
        Field('IpDiscovery'),
        Field('ARN'),
        Field('GlobalReplicationGroupInfo'),
        Field('Description'),
        DictField('NodeGroups'),
        Field('MemberClusters'),
    )


@group.app
class ElasticacheCacheClusterCli(BaseCliApp):
    name: str = 'cachecluster'
    description: str = 'Elasticache cache cluster'

    service_cls: Type[AwsElasticacheCacheClusterService] = (
        AwsElasticacheCacheClusterService)

    stats_fields = ['Region', 'Account', 'Engine']

    short_fields = (
        IDField('CacheClusterId'),
        Field('CacheNodeType'),
        Field('Engine'),
        Field('EngineVersion'),
        Field('CacheClusterStatus'),
        Field('NumCacheNodes'),
    )

    long_fields = (
        Field('CacheSubnetGroupName'),
        Field('ReplicationGroupId'),
        Field('PreferredAvailabilityZone'),
        Field('PreferredMaintenanceWindow'),
        Field('CacheClusterCreateTime'),
        TagField('Tags')
    )

    detail_fields = (
        Field('ARN'),
        BooleanField('AuthTokenEnabled'),
        BooleanField('TransitEncryptionEnabled'),
        BooleanField('AtRestEncryptionEnabled'),
        BooleanField('AutoMinorVersionUpgrade'),
        BooleanField('ReplicationGroupLogDeliveryEnabled'),
        Field('NetworkType'),
        Field('IpDiscovery'),
        Field('SnapshotRetentionLimit'),
        Field('SnapshotWindow'),
        DictField('CacheParameterGroup'),
        Field('LogDeliveryConfigurations'),
        DictField('SecurityGroups'),
        DictField('PendingModifiedValues'),
        Field('CacheSecurityGroups'),
        Field('ClientDownloadLandingPage'),
    )


@group.app
class ElasticacheEventCli(BaseCliApp):
    name: str = 'event'
    description: str = 'Elasticache event'

    service_cls = AwsElasticacheEventService

    short_fields = (
        Field('SourceIdentifier'),
        Field('SourceType'),
        Field('Message'),
        Field('Date'),
    )

    long_fields = ()
    detail_fields = ()



from typing import Any
from itertools import chain

from mypy_boto3_ec2 import EC2Client

from ei.core import defaults
from ei.core.service import BaseAwsService


class AwsEc2VpcService(BaseAwsService):
    service_name = 'ec2'

    @classmethod
    def _list(cls, client: EC2Client) -> Any:
        return client.describe_vpcs()['Vpcs']

    @classmethod
    def _show(cls, client: EC2Client, id: str) -> Any:
        return client.describe_vpcs(VpcIds=[id])['Vpcs']


class AwsEc2SubnetService(BaseAwsService):
    service_name = 'ec2'

    @classmethod
    def _list(cls, client: EC2Client) -> Any:
        return client.describe_subnets()['Subnets']

    @classmethod
    def _show(cls, client: EC2Client, id: str) -> Any:
        return client.describe_subnets(SubnetIds=[id])['Subnets']


class AwsEc2InstanceService(BaseAwsService):
    service_name = 'ec2'

    @classmethod
    def _list(cls, client: EC2Client) -> Any:
        reservations = client.describe_instances()['Reservations']
        instances = [
            reservation['Instances']
            for reservation in reservations
        ]

        iterable = chain(*instances)

        return iterable

    @classmethod
    def _show(cls, client: EC2Client, id: str) -> Any:
        reservations = client.describe_instances(
            InstanceIds=[id]
        )['Reservations']

        instance = [
            reservation['Instances']
            for reservation in reservations
        ][0]

        return instance


class AwsEc2AmiService(BaseAwsService):
    service_name = 'ec2'

    @classmethod
    def _list(cls, client: EC2Client) -> Any:
        images = client.describe_images(
            Owners=defaults.EI_ACCOUNT_IDS)['Images']
        return images

    @classmethod
    def _show(cls, client: EC2Client, id: str) -> Any:
        image = client.describe_images(ImageIds=[id])['Images']

        return image



from typing import Dict
from typing import List

from rich import box
from rich.table import Table


DETAIL_FIELDS = ['key', 'value']


def list_table(fields: List[str], items: List) -> Table:
    table = Table(*fields, box=box.ROUNDED)

    for item in items:
        table.add_row(*[v for v in item])

    return table


def detail_table(item: Dict) -> Table:
    table = Table(*DETAIL_FIELDS, box=box.ROUNDED)

    for key, value in item.items():
        table.add_row(key, value)

    return table




class BaseError(RuntimeError):
    def __str__(self) -> str:
        return f'{self.__class__.__name__}: {"".join(self.args)}'


class PreflightError(BaseError):
    pass


class WrongRegionError(BaseError):
    pass


class WrongAccountError(BaseError):
    pass


class ResourceNotfoundError(BaseError):
    pass



import typing
from typing import Any
from typing import Callable
from typing import Optional
from typing import Collection

from rich.pretty import pretty_repr

if typing.TYPE_CHECKING:
    from typing import List  # noqa: F401


def _serialize(record: Any, raw_value: Any) -> Any:
    return str(raw_value)


def extract(key: str) -> Callable:
    def _serialize(record: Any, raw_value: Any) -> str:
        value = raw_value.get(key, '')
        return str(value)

    return _serialize


def extract_from_tag(key: str) -> Callable:
    def _serialize(record: Any, raw_value: Any) -> str:
        if record.get('Tags'):
            found_tag = [
                tag['Value'] for tag in record['Tags']
                if tag['Key'] == key
            ]  # type: List[str]

            if found_tag:
                return found_tag[0]

        return ''

    return _serialize


class Field(object):
    _name: str
    serialize: Callable[[Any, Any], str]

    def __init__(self, name: str,
                 serializer: Optional[Callable] = None) -> None:
        self._name = name
        if serializer:
            self.serialize = serializer
        else:
            if not hasattr(self, 'serialize'):
                self.serialize = _serialize

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__name__}@{self.__hash__()}: "{self._name}" >'
        )

    def __call__(self, record: Any) -> str:
        raw_value = record.get(self._name, '')

        return self.serialize(record, raw_value)


class IDField(Field):
    def serialize(self, record: Any, raw_value: Any) -> str:
        return f'[bold]{raw_value}[/bold]'


class BooleanField(Field):
    def serialize(self, record: Any, raw_value: Any) -> str:
        color = 'green' if raw_value is True else 'bright_red'
        return f'[{color}]{raw_value}[/{color}]'


class TagField(Field):
    def serialize(self, record: Any, raw_value: Any) -> str:
        if not raw_value:
            return ''

        return '\n'.join([
            f'[bright_black]{tag["Key"]}[/bright_black]: {tag["Value"]}'
            for tag in raw_value
        ])


class DictField(Field):
    def _render_obj(self, obj: dict) -> str:
        if not obj:
            return '[bright_black]None[/bright_black]'

        return '\n'.join([
            f'[bright_black]{key}[/bright_black]: '
            f'{pretty_repr(value) if isinstance(value, Collection) else value}'
            for key, value in obj.items() if key != 'ResponseMetadata'
        ])

    def serialize(self, record: Any, raw_value: Any) -> str:
        if type(raw_value) is list:
            return '\n'.join([
                '\n'.join([
                    (('- ' if idx == 0 else '  ') + line)
                    for idx, line in enumerate(
                        self._render_obj(obj).split('\n'))
                ])
                for obj in raw_value
            ])
        else:
            return self._render_obj(raw_value)



from unittest import mock

from tests.mocks.app import DummyAwsService


@mock.patch('ei.core.defaults.AWS_REGION', 'region-1')
@mock.patch('ei.core.defaults.EI_REGIONS', ['region-1', 'region-2'])
@mock.patch('ei.core.defaults.EI_ACCOUNT_IDS', ['account-a', 'account-b'])
def test_dummy_service() -> None:
    service = DummyAwsService()

    assert service.service_name == 'dummy'
    items = service.list(region='test-region', account_id='test-account')
    assert items[0]['Region'] == 'test-region'
    assert items[0]['Account'] == 'test-account'

    item = service.show(
        id='item-a', region='test-region', account_id='test-account')
    assert item['Region'] == 'test-region'
    assert item['Account'] == 'test-account'

    item = service.show(
        id='item-a', region='', account_id='test-account')
    assert item['Region'] == 'region-1'
    assert item['Account'] == 'test-account'

    item = service.show(
        id='item-a', region='', account_id='')
    assert item['Region'] == 'region-1'
    assert item['Account'] == 'account-a'



from typing import Any
from unittest import mock

from tests.mocks.app import DummyAwsCli


@mock.patch('ei.core.cli.defaults')
def test_cli_app_should_returns_stats(mock_defaults: Any) -> None:
    mock_defaults.AWS_REGION = 'test-region'
    mock_defaults.AWS_ACCESS_KEY_ID = 'test-access-key'
    mock_defaults.AWS_SECRET_ACCESS_KEY = 'test-secret-access-key'
    mock_defaults.AWS_SECURITY_TOKEN = 'test-security-token'
    mock_defaults.AWS_SESSION_EXPIRATION = 'test-expiration'

    app = DummyAwsCli()
    app.list()
    app.show(id='item-a')



# from typing import Type

from ei.core.cli import BaseCliApp
from ei.core.cli import CliGroup
from ei.core.fields import Field
from ei.core.fields import IDField
from ei.core.fields import BooleanField
from ei.core.fields import TagField
from ei.core.fields import DictField
from ei.services.aws.rds import AwsRdsInstanceService


group = CliGroup(name='rds', description='AWS RDS')


@group.app
class Ec2RdsInstanceCli(BaseCliApp):
    name: str = 'instance'
    description: str = 'EC2 RDS'

    service_cls = AwsRdsInstanceService

    stats_fields = ['Region', 'Account', 'Engine']

    short_fields = (
        IDField('DBInstanceIdentifier'),
        Field('DBInstanceClass'),
        Field('Engine'),
        Field('EngineVersion'),
        Field('DBInstanceStatus'),
    )

    long_fields = (
        BooleanField('PubliclyAccessible'),
        BooleanField('StorageEncrypted'),
    )

    detail_fields = (
        Field('AvailabilityZone'),
        Field('AutomaticRestartTime'),
        Field('MasterUsername'),
        Field('DBName'),
        Field('AllocatedStorage'),
        Field('InstanceCreateTime'),
        Field('PreferredBackupWindow'),
        Field('BackupRetentionPeriod'),
        Field('PreferredMaintenanceWindow'),
        Field('LatestRestorableTime'),
        BooleanField('MultiAZ'),
        BooleanField('AutoMinorVersionUpgrade'),
        Field('ReadReplicaSourceDBInstanceIdentifier'),
        Field('ReplicaMode'),
        Field('LicenseModel'),
        Field('Iops'),
        Field('CharacterSetName'),
        Field('NcharCharacterSetName'),
        Field('SecondaryAvailabilityZone'),
        Field('StorageType'),
        Field('TdeCredentialArn'),
        Field('DbInstancePort'),
        Field('DBClusterIdentifier'),

        Field('KmsKeyId'),
        Field('DbiResourceId'),
        Field('CACertificateIdentifier'),
        BooleanField('CopyTagsToSnapshot'),
        Field('MonitoringInterval'),
        Field('EnhancedMonitoringResourceArn'),
        Field('MonitoringRoleArn'),
        Field('PromotionTier'),
        Field('DBInstanceArn'),
        Field('Timezone'),
        BooleanField('IAMDatabaseAuthenticationEnabled'),
        BooleanField('PerformanceInsightsEnabled'),
        Field('PerformanceInsightsKMSKeyId'),
        Field('PerformanceInsightsRetentionPeriod'),
        BooleanField('DeletionProtection'),
        Field('MaxAllocatedStorage'),
        BooleanField('CustomerOwnedIpEnabled'),
        Field('AwsBackupRecoveryPointArn'),
        Field('ActivityStreamStatus'),
        Field('ActivityStreamKmsKeyId'),
        Field('ActivityStreamKinesisStreamName'),
        Field('ActivityStreamMode'),
        Field('ActivityStreamEngineNativeAuditFieldsIncluded'),
        Field('AutomationMode'),
        Field('ResumeFullAutomationModeTime'),
        Field('CustomIamInstanceProfile'),
        Field('BackupTarget'),
        Field('NetworkType'),
        Field('ActivityStreamPolicyStatus'),
        Field('StorageThroughput'),
        Field('DBSystemId'),
        DictField('Endpoint'),
        DictField('DBSecurityGroups'),
        DictField('VpcSecurityGroups'),
        DictField('DBParameterGroups'),
        DictField('DBSubnetGroup'),
        DictField('PendingModifiedValues'),
        DictField('OptionGroupMemberships'),
        DictField('DomainMemberships'),
        DictField('AssociatedRoles'),
        Field('ReadReplicaDBInstanceIdentifiers'),
        Field('ReadReplicaDBClusterIdentifiers'),
        Field('StatusInfos'),
        Field('EnabledCloudwatchLogsExports'),
        Field('ProcessorFeatures'),
        Field('ListenerEndpoint'),
        Field('DBInstanceAutomatedBackupsReplications'),
        Field('MasterUserSecret'),
        Field('CertificateDetails'),
        TagField('TagList'),
    )



# from typing import Type

from ei.core.cli import BaseCliApp
from ei.core.cli import CliGroup
from ei.core.fields import Field
from ei.core.fields import IDField
from ei.core.fields import DictField
from ei.core.fields import TagField
from ei.services.aws.s3 import AwsS3BucketService


group = CliGroup(name='s3', description='AWS S3')


@group.app
class S3BucketCli(BaseCliApp):
    name: str = 'bucket'
    description: str = 'S3 Bucket'

    service_cls = AwsS3BucketService

    stats_fields = ['Region', 'Account']

    short_fields = (
        IDField('Name'),
        Field('CreationDate'),
    )

    long_fields = ()
    detail_fields = (
        DictField('AccelerateConfiguration'),
        DictField('Acl'),
        DictField('Encryption'),
        DictField('Location'),
        DictField('Logging'),
        DictField('Notification'),
        DictField('NotificationConfiguration'),
        DictField('Policy'),
        DictField('PolicyStatus'),
        DictField('RequestPayment'),
        DictField('Versioning'),
        DictField('Cors'),
        DictField('Lifecycle'),
        DictField('LifecycleConfiguration'),
        DictField('OwnershipControls'),
        DictField('Replication'),
        DictField('Website'),
        DictField('AnalyticsConfigurations'),
        DictField('IntelligentTieringConfigurations'),
        DictField('InventoryConfigurations'),
        DictField('MetricsConfigurations'),
        TagField('Tagging'),
    )



import datetime
from contextlib import contextmanager
from typing import TYPE_CHECKING
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Optional

import boto3
from botocore.credentials import DeferredRefreshableCredentials
from botocore.session import get_session

from ei.core import defaults


if TYPE_CHECKING:
    from botocore.session import Session  # noqa: F401


def from_sts(account_id: str) -> Dict:
    sts_client = boto3.Session().client('sts')

    role_arn = defaults.EI_STS_ASSUME_ROLE_ARN_PATTERN.format(
        account_id=account_id)

    assumed_role_object: dict = sts_client.assume_role(
        RoleArn=role_arn,
        RoleSessionName=defaults.EI_STS_ASSUME_ROLE_SESSION_NAME)

    credentials: dict = assumed_role_object['Credentials']

    return credentials


def from_env(*args: tuple, **kwargs: dict) -> Dict:
    credentials = {
        'AccessKeyId': defaults.AWS_ACCESS_KEY_ID,
        'SecretAccessKey': defaults.AWS_SECRET_ACCESS_KEY,
        'SessionToken': defaults.AWS_SECURITY_TOKEN,
        'Expiration': datetime.datetime.fromisoformat(
            defaults.AWS_SESSION_EXPIRATION)
    }

    return credentials


credential_resolvers = {
    'env': from_env,
    'sts': from_sts
}  # type: Dict[str, Callable]


def create_session(
        service_name: str, account_id: str,
        region: Optional[str] = defaults.AWS_REGION) -> boto3.Session.client:
    """Creates client session via sts client
    """

    credential_resolver: Callable = credential_resolvers[
        defaults.EI_CREDENTIAL_RESOLVER]

    def _get_session_creds() -> Dict:
        credentials = credential_resolver(account_id)

        credential_metadata = {
            'access_key': credentials['AccessKeyId'],
            'secret_key': credentials['SecretAccessKey'],
            'token': credentials['SessionToken'],
            'expiry_time': credentials['Expiration'].isoformat(),
        }

        return credential_metadata

    credentials = DeferredRefreshableCredentials(
        refresh_using=_get_session_creds, method='sts-assume-role')

    session = get_session()  # type: Session
    session.set_config_variable('region', region)
    session._credentials = credentials

    session = boto3.Session(botocore_session=session)

    return session.client(service_name)


@contextmanager
def client_session(
        service_name: str, account_id: str,
        region: Optional[str] = defaults.AWS_REGION) -> Generator:
    session = create_session(
        service_name=service_name, account_id=account_id, region=region)

    yield session




from ei.core.cli import BaseCliApp
from ei.core.cli import CliGroup
from ei.core.fields import Field
from ei.core.fields import IDField
from ei.core.fields import TagField
from ei.core.fields import BooleanField
from ei.core.fields import DictField
from ei.core.fields import extract
from ei.services.aws.elb import AwsElbLoadbalancerService
from ei.services.aws.elb import AwsElbListenerService
from ei.services.aws.elb import AwsElbTargetGroupService


group = CliGroup(name='elb', description='AWS ElasticLoadbalancer')


@group.app
class ElbLoadbalancerCli(BaseCliApp):
    name: str = 'loadbalancer'
    description: str = 'ElasticLoadbalancer Loadbalander'

    service_cls = AwsElbLoadbalancerService

    stats_fields = ['Region', 'Account', 'Type']

    short_fields = (
        IDField('LoadBalancerArn'),
        Field('LoadBalancerName'),
        Field('State', serializer=extract('Code')),
        Field('Type'),
    )

    long_fields = (
        Field('IpAddressType'),
        Field('Scheme'),
        Field('CreatedTime'),
        Field('CanonicalHostedZoneId'),
    )

    detail_fields = (
        Field('VpcId'),
        Field('DNSName'),
        DictField('Attributes'),
        DictField('AvailabilityZones'),
        DictField('Listeners'),
        DictField('TargetGroups'),
        TagField('Tags')
    )


@group.app
class ElbListenerCli(BaseCliApp):
    name: str = 'listener'
    description: str = 'ElasticLoadbalancer Listener'

    service_cls = AwsElbListenerService

    short_fields = (
        IDField('ListenerArn'),
        Field('LoadBalancerArn'),
        Field('Port'),
        Field('Protocol'),
    )

    long_fields = ()

    detail_fields = (
        DictField('DefaultActions'),
        DictField('Certificates'),
        DictField('Rules'),
        TagField('Tags')
    )


@group.app
class ElbTargetGroupCli(BaseCliApp):
    name: str = 'targetgroup'
    description: str = 'ElasticLoadbalancer TargetGroup'

    service_cls = AwsElbTargetGroupService

    short_fields = (
        IDField('TargetGroupArn'),
        Field('TargetGroupName'),
        Field('Protocol'),
        Field('Port'),
        BooleanField('HealthCheckEnabled'),
        Field('TargetType'),
    )

    long_fields = (
        Field('HealthCheckProtocol'),
        Field('HealthCheckPort'),
        Field('HealthCheckIntervalSeconds'),
        Field('HealthCheckTimeoutSeconds'),
        Field('HealthyThresholdCount'),
        Field('UnhealthyThresholdCount'),
        Field('HealthCheckPath'),
    )

    detail_fields = (
        Field('VpcId'),
        Field('ProtocolVersion'),
        Field('IpAddressType'),
        DictField('Matcher'),
        Field('LoadBalancerArns'),
        DictField('DefaultActions'),
        DictField('Certificates'),
        DictField('Rules'),
        TagField('Tags')
    )



import json
from typing import Any

from mypy_boto3_s3 import S3Client
from ei.core.exceptions import ResourceNotfoundError
from ei.core.service import BaseAwsService
from ei.core.utils import to_pascal_case


class BucketNotFoundError(ResourceNotfoundError):
    pass


S3_PROPERTY_FETCHER = {
    'get': (
        'accelerate_configuration',
        'acl',
        'encryption',
        'location',
        'logging',
        'notification',
        'notification_configuration',
        'policy',
        'policy_status',
        'request_payment',
        'tagging',
        'versioning',
        'cors',
        'lifecycle',
        'lifecycle_configuration',
        'ownership_controls',
        'replication',
        'website',
    ),
    'list': (
        'analytics_configurations',
        'intelligent_tiering_configurations',
        'inventory_configurations',
        'metrics_configurations',
    ),
}


class AwsS3BucketService(BaseAwsService):
    service_name = 's3'

    @classmethod
    def _list(cls, client: S3Client) -> Any:
        return client.list_buckets()['Buckets']

    @classmethod
    def _show(cls, client: S3Client, id: str) -> Any:
        bucket_list = cls._list(client)
        found_bucket = [b for b in bucket_list if b['Name'] == id]
        if not found_bucket:
            raise BucketNotFoundError(id)

        bucket = {**found_bucket[0]}
        for method, properties in S3_PROPERTY_FETCHER.items():
            for property in properties:
                fetcher = getattr(client, f'{method}_bucket_{property}')

                try:
                    result = fetcher(Bucket=id)
                    bucket[to_pascal_case(property)] = {**result}
                except Exception:
                    pass

        bucket['Tagging'] = bucket['Tagging']['TagSet']
        bucket['Policy'] = json.loads(bucket['Policy'].get('Policy', '{}'))

        return [bucket]
